#include <easy.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>

e_obj_descriptor e_obj_descriptor_file = {
  "file",             /* Name      */
  E_OBJ_MODE_NO_CLONE,/* def_mode  */
  malloc,             /* _alloc    */
  free,               /* _free     */
  realloc,            /* _realloc  */
  e_file_free,        /* free      */
  e_file_to_string,   /* to_string */
  e_file_clone,       /* clone     */
  e_file_equals,      /* equals    */
};


/************************************************************/
/* FILE :                                                   */
/* Files are objects to refer to open files.                */
/* The object is seen as a reference to a file as opposed   */
/* to its content. This means that comparing 2 files checks */
/* if the the two e_files refer to the same file open       */
/* with the same mode. Converting an e_file to a string     */
/* yields something like Open("Filename",MODE)              */
/***********************************************************/
/* Typical usage                                           */
/*                                                         */
/* char c;                                                 */
/* e_file *fp;                                             */
/*  E_NULL_ERROR(fp = e_file_new(0,0,"FileName","r"));     */
/*  while((c = fget(E_FILE_FPT(fp)))!= EOF) { ... }        */
/*  E_FREE(fp);      //   Close and free the file          */
/*                                                         */
/* char c;                                                 */
/* e_file *fp;                                             */
/* e_string *sp;                                           */
/*  E_NULL_ERROR(sp = e_string_new(0,"FileName,0"));       */
/*  E_NULL_ERROR(fp = e_file_new(0,sp,0,"w"));             */
/*  while((c = fget(E_FILE_FPT(fp)))!= EOF) { ... }        */
/*  E_FREE(fp); // Close anf free the file                 */
/***********************************************************/
/***************************************************************************/
/* Create or initialise a new e_file                                       */
/* fp : e_file to initialise (static e_file)                               */
/*      if NULL allocate it                                                */
/*      fp is assumed to be of the correct type                            */
/* fn : e_string file name. If null, file_name is used                     */ 
/*      "" or "-" => open stdout (mode r) or stdout (mode w and a )        */
/* file_name : constant filename (ignored if fn is non null)               */
/* mode : constant string. Same mode as fopen.                             */
/* Return e_file                                                           */
/*        0 : error                                                        */
/* Link : fn -1                                                            */
/*        Return +1                                                        */
/***************************************************************************/
e_file *e_file_new(e_file *fp,e_string *fn,char *file_name,char *mode)
{ 

                 /* Static object or allocate it ? */
  if (!fp)
  { if (!(fp = e_obj_descriptor_file._alloc(sizeof(e_file)))) 
      return((e_file *)e_error_no_memory("e_file_new"));
    fp->h.refs = 1;
  }
  else fp->h.refs = -1;
  fp->h.mode = 0;

  fp->h.obj_desc_p = &e_obj_descriptor_file;

                 /* is mode non zero ? */
  if(!mode) 
  { e_error_null_pointer("e_file_new(-,-,-,mode)");
    goto error;
  }

  fp->file_name = 0;

  if(!fn)        /* Allocate the file name */
  { if(!file_name) 
    { e_error_null_pointer("e_file_new(-,file_name,-,-)");
      goto error;
    }
    E_NULL_ERROR(fp->file_name = e_string_new(0,file_name,0));
    fn = fp->file_name;
  }
  else        /* Use the e_string file name */
  { fp->file_name = fn;
  }

  fp->mode = mode;


             /* "" or "-" : open stdi or stdout */
  if((E_STRING_S(fp->file_name)[0] == 0) || 
      ((E_STRING_S(fp->file_name)[0] == '-') && 
       (E_STRING_S(fp->file_name)[1] == 0)
      )
    )
  { if(E_FILE_MODE(fp)[0] == 'r') fp->fpt = stdin;
    else fp->fpt = stdout;
  }
  else if(!(fp->fpt = fopen(E_STRING_S(fp->file_name),mode)))
  { e_error(E_ERROR_FILE,E_ERROR_LEVEL_USER,
            "Unable to open file %s with mode %s",
            E_STRING_S(fp->file_name),mode);
    goto error;
  }

  return(fp);

error:
  if(fn) E_FREE(fn);
  if(fp && (fp->h.refs > 0)) e_obj_descriptor_file._free(fp);
  return(0); 
}

/***************************************************************************/
/* Free and close the file                                                 */
/* If the refs counter is -1 : static file : close the file and            */
/*                             free the file name.                         */
/* If the refs counter is 1 : free the allocated memory for the e_file    */
/* Otherwise decrement the counter by 1.                                   */
/* REF : stp : -1                                                          */
/***************************************************************************/
void e_file_free(void *fp)
{

  if(fp)
  { if(E_OBJ_REF(fp) == 0)
    { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
            "Trying to free a free file (%x) ",fp);
      goto error;
    }

    if (E_OBJ_REF(fp) < 0)      /* static file, free the file_name */
    { if (E_FILE_FPT(fp)) fclose(E_FILE_FPT(fp));
      if (E_FILE_FILENAME(fp)) 
          e_obj_descriptor_file._free(E_FILE_FILENAME(fp));
      E_FILE_FILENAME(fp) = 0; 
      return;
    }

                                /* decrement the refs counter   */
    if (E_OBJ(fp)->refs-- > 1) return;

                                /* Last reference, 
                                   free the file filename  */
    if (E_FILE_FPT(fp)) fclose(E_FILE_FPT(fp));
    if (E_FILE_FILENAME(fp)) e_obj_descriptor_file._free(E_FILE_FILENAME(fp));

                                /* free the e_file */
    e_obj_descriptor_file._free(fp);
  }
  return;

error:
  e_error_flush(E_ERROR_LEVEL_DEBUG); 
  
}

/*************************************************************/
/* Convert the file to a string                              */
/* The file is converted to the string                       */
/*   Open(file_name,mode)                                    */
/* fp : a file, assumed to be of the correct type            */
/* tab: left identation (ignored for COMPACT mode)           */
/* mode:  0, E_TO_STR_COMPACT                                */
/* Return : new string                                       */
/*          0 : error                                        */
/* LINK : fp : 0                                             */
/*        RETURN : +1                                        */
/*************************************************************/
void *e_file_to_string(void *fp,int tab,e_flag mode)
{ e_string *stp = 0;

  E_NULL_ERROR(stp = e_string_new(0,"Open(",
                            E_STRING_LEN(E_FILE_FILENAME(fp))+15));

  E_NULL_ERROR(stp = e_string_append_buffer(stp,
                                  E_STRING_S(E_FILE_FILENAME(fp))));
  E_NULL_ERROR(stp = e_string_append_buffer(stp,",\""));
  E_NULL_ERROR(stp = e_string_append_buffer(stp,E_FILE_MODE(fp)));
  E_NULL_ERROR(stp = e_string_append_buffer(stp,"\")"));


  if(!(mode & E_TO_STR_NO_TAB))
  { /* add tab to the string */
    stp = e_string_insert_chars(stp,0,tab,' ');
  }

  return (stp);

error:
  if (stp) E_FREE(stp);
  return(0);
}

/*****************************************************************/
/* Clone a file                                                  */
/* mode : cloning mode                                           */
/* If the file can't be cloned (default), return fp              */
/* and increase its refs counter                                 */
/* If cloned, the file is just reopened with the same file mode. */
/* The seek position is thus determined by the file mode         */
/* fp : a file, assumed to be of the correct type                */
/* Return : cloned e_file                                        */
/*          0 : Error                                            */
/* LINK : stp : 0                                                */
/*        RETURN : +1                                            */
/*****************************************************************/
void *e_file_clone(void *fp,e_flag mode)
{ e_file *newfp = 0;

                 /* Can't clone : return the same object */
  if(E_OBJ_NO_CLONE(fp)) return(E_USE(fp));

  E_NULL_ERROR(newfp = e_file_new(0,E_FILE_FILENAME(fp),0,E_FILE_MODE(fp)));
  
  return(newfp);

error:
  if(newfp) E_FREE(newfp); 
  return(0);
}

/*************************************************************/
/* Test if the 2 files are equal:                            */
/* The two files are equal if they are identical (p1 == p2), */
/* or if they have the same file_name and the same mode.     */
/* p1 is assumed to be of type e_file                        */
/* This does not compare the seek position of the file!!!    */
/* Return : 1 : p1 and p2 are equal                          */
/*          0 : p1 and p2 are different                      */
/* Link : p1 0                                               */
/*        p2 0                                               */
/*************************************************************/
int e_file_equals(void *p1,void *p2)
{ 

  if (p1 == p2)   return(1);
  if (!p1 || !p2) return(0);

            /* different types => differ */
  if(E_OBJ_DESCR(p1) != E_OBJ_DESCR(p2)) return(0);

            /* different modes => differ */
  if(strcmp(E_FILE_MODE(p1),E_FILE_MODE(p1))) return(0);

            /* then compare the file names */
  return(E_EQUALS(E_FILE_FILENAME(p1),E_FILE_FILENAME(p2)));

}

/**************************************************/
/* Check if a file exists and return its st_mode. */
/* Return : 0 : file does not exist.              */
/*          st_mode : file exist                  */
/**************************************************/
int e_stat_file(char *filename)
{ struct stat statp;

  if(!filename) return(0);

  if(!stat(filename,&statp))
  { return(statp.st_mode);
  }
  else return(0);
}

/**************************************************/
/* Check if a file exists and open it             */
/* Return : FILE POINTER                          */
/*          0 : error                             */
/**************************************************/
FILE *e_open_file(char *file_name,char *mode)
{ FILE *fpt;
  
  if(!strcmp(mode,"r"))
  { if(!S_ISREG(e_stat_file(file_name)) || !(fpt = fopen(file_name,"r")))
    { e_error(E_ERROR_FILE,E_ERROR_LEVEL_USER,
              "Can't open file %s",file_name);
      return(0);
    }
  }
  else
  { if(!(fpt = fopen(file_name,mode)))
    { e_error(E_ERROR_FILE,E_ERROR_LEVEL_USER,
              "Can't open file %s",file_name);
      return(0);
    }
  }

  return(fpt);
}
