#include <easy.h>
#include <stdlib.h>

e_obj_descriptor e_obj_descriptor_list = {
  "list",             /* Name      */
  0,                  /* def_mode  */
  malloc,             /* _alloc    */
  free,               /* _free     */
  realloc,            /* _realloc  */
  e_list_free,        /* free      */
  e_list_to_string,   /* to_string */
  e_list_clone,       /* clone     */
  e_list_equals,      /* equals    */
};


/************************************************************/
/* List :                                                   */
/* Binary tree or list                                      */
/************************************************************/

/***************************************************************************/
/* Create or initialise a new e_list                                       */
/* lp : e_list to initialise (static e_list)                               */
/*      if NULL allocate it                                                */
/*      lp is assumed to be of the correct type                            */
/* val : list value : if null, empty element                               */
/* next : next element in the list. If null => end of list                 */
/* Return e_list                                                           */
/*        0 : error                                                        */
/* Link : np -1                                                            */
/*        val -1   : passed on to the list                                 */
/*        next -1  : passed on to the list                                 */
/*        Return +1                                                        */
/***************************************************************************/
e_list *e_list_new(e_list *lp,void *val,e_list *next)
{ 
                 /* Static object or allocate it ? */
  if (!lp)
  { if (!(lp = e_obj_descriptor_list._alloc(sizeof(e_list)))) 
      return((e_list *)e_error_no_memory("e_list_new"));
    lp->h.refs = 1;
  }
  else lp->h.refs = -1;
  lp->h.mode = 0;

  lp->h.obj_desc_p = &e_obj_descriptor_list;

  E_NULL_ERROR(e_check_type(next,E_TYPE_LIST,"e_list_new(-,-,next)"));

  E_LIST_VAL(lp) = val;  
  E_LIST_NEXT(lp) = next;  

  return(lp);

error:
  if(val) E_FREE(val);
  if(next) E_FREE(next);
  if(lp && (lp->h.refs > 0)) e_obj_descriptor_list._free(lp);
  return(0); 
}

/***************************************************************************/
/* Free the e_list                                                         */
/* If the refs counter is -1 : static list : leave as is                   */
/* If the refs counter is 1 : free the allocated memory for the e_list     */
/* Otherwise decrement the counter by 1.                                   */
/* REF : stp : -1                                                          */
/***************************************************************************/
void e_list_free(void *lp)
{

  if(lp)
  { if(E_OBJ_REF(lp) == 0)
    { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
            "Trying to free a free list (%x) ",lp);
      goto error;
    }
                                /* static list */
    if (E_OBJ_REF(lp) < 0) return;

                                /* decrement the refs counter   */
    if (E_OBJ(lp)->refs-- > 1) return;

                                /* Last reference,     
                                   Free the two element, 
                                    then free the list */
    if(E_LIST_VAL(lp)) E_FREE(E_LIST_VAL(lp));
    if(E_LIST_NEXT(lp)) E_FREE(E_LIST_NEXT(lp));

    e_obj_descriptor_list._free(lp);
  }
  return;

error:
  e_error_flush(E_ERROR_LEVEL_DEBUG); 
  
}

/*************************************************************/
/* Convert the list to a string                              */
/* lp : an e_list, assumed to be of the correct type         */
/* Return : new string                                       */
/*          0 : error                                        */
/* LINK : lp : 0                                             */
/*        RETURN : +1                                        */
/*************************************************************/
void *e_list_to_string(void *lp,int tab,e_flag mode)
{ e_string *stp = 0;
  char *s;
  
                             /* only write a ( before the first element */
  if(mode & E_TO_STR_NOT_FIRST) s = " ";
  else s = "(";

  E_NULL_ERROR(stp = e_string_new(0,s,E_LINE_SIZE));
  
                             /* Value */
  if(E_LIST_VAL(lp))
  { E_NULL_ERROR(stp = e_string_append_string(stp,
              E_TO_STRING(E_LIST_VAL(lp),0,E_TO_STR_EMBEDED)));
  }

                             /* Rest of list */
  if(E_LIST_NEXT(lp)) 
  { E_NULL_ERROR(stp = e_string_append_string(stp,
              E_TO_STRING(E_LIST_NEXT(lp),0,
                          E_TO_STR_EMBEDED|E_TO_STR_NOT_FIRST)));
  }
  else                       /* End of list */
  { E_NULL_ERROR(stp = e_string_append_buffer(stp,")"));
  }

  if(!(mode & E_TO_STR_NO_TAB))
  { /* add tab to the string */
    stp = e_string_insert_chars(stp,0,tab,' ');
  }

  return (stp);

error:
  if (stp) E_FREE(stp);
  return(0);
}

/*****************************************************************/
/* Clone a list                                                  */
/* mode : cloning mode                                           */
/* If the list can't be cloned (default), return lp              */
/* and increase its refs counter                                 */
/* lp : a list, assumed to be of the correct type                */
/* Return : cloned e_list                                        */
/*          0 : Error                                            */
/* LINK : stp : 0                                                */
/*        RETURN : +1                                            */
/*****************************************************************/
void *e_list_clone(void *lp,e_flag mode)
{ e_list *next,*newlp;
  void *val;

                 /* Can't clone : return the same object */
  if(E_OBJ_NO_CLONE(lp)) return(E_USE(lp));

  val = next = 0;

  if(mode & E_CLONE_RECURSIVE)
  {                      /* clone recursively */   
    if(E_LIST_VAL(lp))  { E_NULL_ERROR(val  = E_CLONE(E_LIST_VAL(lp),mode));  }
    if(E_LIST_NEXT(lp)) { E_NULL_ERROR(next = E_CLONE(E_LIST_NEXT(lp),mode)); }
    
  }
  else                   /* only clone the top list element */
  { if(E_LIST_VAL(lp))  val =  E_USE(E_LIST_VAL(lp));
    if(E_LIST_NEXT(lp)) next = E_USE(E_LIST_NEXT(lp));
  }

  E_NULL_ERROR(newlp = e_list_new(0,val,next));

  return(newlp);

error:
  if(val) E_FREE(val); 
  if(next) E_FREE(next); 
  return(0);
}

/***************************************************************/
/* Test if the 2 lists are equal:                              */
/* The two lists are equal if they are identical (p1 == p2),   */
/* or if they have the same value and same tail list (nexts)   */
/* p1 is assumed to be of type e_list                          */
/* Return : 1 : p1 and p2 are equal                            */
/*          0 : p1 and p2 are different                        */
/* Link : p1 0                                                 */
/*        p2 0                                                 */
/***************************************************************/
int e_list_equals(void *p1,void *p2)
{ 

  if (p1 == p2)   return(1);
  if (!p1 || !p2) return(0);

            /* different objet  types => differ */
  if(E_OBJ_DESCR(p1) != E_OBJ_DESCR(p2)) return(0);

            /* compare the values */
  if(E_LIST_VAL(p1) != E_LIST_VAL(p2))
  { if ((E_LIST_VAL(p1)==0) || (E_LIST_VAL(p2)==0) ||
        !E_EQUALS(E_LIST_VAL(p1),E_LIST_VAL(p2))) 
       return(0);
  }

            /* compare the list tails */
  if(E_LIST_NEXT(p1) != E_LIST_NEXT(p2))
  { if ((E_LIST_NEXT(p1)==0) || (E_LIST_NEXT(p2)==0) ||
        !E_EQUALS(E_LIST_NEXT(p1),E_LIST_NEXT(p2))) 
       return(0);
  }
  return(1);

}
