#include <easy.h>

extern char VERSION[];

int e_parameter_Set_Default_Value(e_parameter *pars);
e_parameter *e_parameter_Get_Par_Details(e_parameter *pars,char *name);
int e_parameter_Check_Par_Read(e_parameter *pars);


/***************************************************************************/
/* Read the parameters from a comand argument                              */
/* Parameters are preceded with the - symbol. They can be of the following */
/* Types : Integer, Real, Char, String, Boolean, and Flag                  */
/* All the parameters are described in the Par array passed to             */
/* e_parameter_read_arg_line e_GetArgs.                                    */
/* Arguments which are not parameters are put in the Argv list.            */
/* -help and -version display an help message and the program version      */
/*  message respectively. -h and -v can be used as aliases by using the    */
/*  E_PARAMETER_H_IS_HELP and E_PARAMETER_V_IS_VERSION modes.              */
/*                                                                         */
/*                                                                         */
/* Example of Parameter array:                                             */
/* NAME TYPE VAR_POINTER DEFAULT READ ALIAS1 ALIAS2                        */
/*  -NAME the argument name (without the -)                                */
/*  -TYPE the type of variable E_CTYPE_INTEGER, E_CTYPE_REAL, E_CTYPE_CHAR */
/*        E_CTYPE_STRING, E_CTYPE_BOOLEAN, E_CTYPE_FLAG,                   */
/* VAR_POINTER : pointer to the C variable to initialise                   */
/* DEFAULT : Default value entered as a string                             */
/*           The string is converted to the correct type.                  */
/*           FLAGS : always OFF by default, => can be set to NULL          */
/* READ : nomaly set to 0, unless it is initialised in the code.           */
/*        Can be set to 2 to avoid error and detect the need to            */
/*        initialise in the code.                                          */
/*        READ is set to 1 when a parameter is set to a value.             */
/* ALIAS1, ALIAS2 two alias names. ("" or 0 if none).                      */
/*                                                                         */
/* Example:                                                                */
/* struct  {                                                               */
/*   int flag;                                                             */
/*   char File[256];                                                       */
/*   int n;                                                                */
/*   double r;                                                             */
/*   int b;                                                                */
/* } par;                                                                  */
/*                                                                         */
/* e_parameter Pars[] = {                                                  */
/* { "short",E_CTYPE_FLAG,&par.flag,0,0,"s",0},                            */
/* { "file",E_CTYPE_STRING,par.File,"Def_File",0,"f",0},                   */
/* { "No",E_CTYPE_INTEGER,&par.n,"10",0,"no","n"},                         */
/* { "Radius",E_CTYPE_REAL,&par.r,"1.2",0,"r","R"},                        */
/* { "Bool",E_CTYPE_REAL,&par.b,"1",0,"b",0},                              */
/* { 0,0,0,0,0,0},                                                         */
/* };                                                                      */
/* char *Help[] = {                                                        */
/* " comm [-short=s] [-file=f FILE_NAME] [-No=no=n N] [Radius=r=R VAL]     */
/*        [-Bool=b BOOL] src",                                             */
/* " comm -help",                                                          */ 
/* 0                                                                       */
/* };                                                                      */
/* char Version[] = "comm: v1.1; 15/10/99;  B. Piette";                    */
/*                                                                         */
/* main(int argc, char **argv)                                             */
/* { char **Argv;                                                          */
/*   int Argc;                                                             */
/*  if((Argc=e_parameter_read_arg_line(argc,argv,Pars,Help,&Argv,          */
/*                     E_PARAMETER_H_IS_HELP|E_PARAMETER_V_IS_VERSION))<0) */
/*     exit(0);                                                            */
/*                                                                         */
/* argc, argv : the argument list                                          */
/* Par the array of Parameter descrbing all the parameters                 */
/* help : char array : a help message displayed when the -help flag is on  */
/* Argvp : pointer array of string pointers used to store the extra        */
/*        arguments e_parameter_read_arg_line allocate it and staves it in */
/*        *Argvp. The last element of Argvp is always NULL.                */
/*        *Argvp can be freed with the <stdlib> free().                    */
/* mode :                                                                  */
/*  E_PARAMETER_NOBADPAR : (0) print an error message for each unknown par.*/
/*  E_PARAMETER_BADPAROK : bad parameters are added to the argument list   */
/*  E_PARAMETER_MINUSOK  : - is added to the argument list (for example,   */
/*                         when used to indicate stdin or stdout.)         */
/*  E_PARAMETER_MINUSLAST : - force all the remaining argument to be added */
/*                      to the argument list.                              */
/*  E_PARAMETER_H_IS_HELP : -h is alias for -help.                         */
/*  E_PARAMETER_V_IS_VERSION : -v is alias for -version.                   */
/*                                                                         */
/* Return : Number of extra argument read (number of entries in *Argv.)    */
/*          -1 : error                                                     */
/***************************************************************************/
int e_parameter_read_arg_line(int argc,char **argv,e_parameter *Par,
               char **help,char ***Argvp,int mode)
{ int i,j,getpars,done;
  e_parameter *par;
  char *s;
  char **Argv;
              

      /* Check for -help, -h,  -version or -v and display help or version */
  done = 0;                  
  for(i=1; i < argc; i++)
  { if((!strcmp("-help",argv[i])) || 
       ((mode & E_PARAMETER_H_IS_HELP) &&(!strcmp("-h",argv[i])))
      )
    { e_Syntax(help);              /* Display help message and exit */
      done = 1;
    }
    if((!strcmp("-version",argv[i])) || 
       ((mode & E_PARAMETER_V_IS_VERSION) &&(!strcmp("-v",argv[i])))
      ) 
    { puts(VERSION);
      done = 1;
    }
  }
  if(done) exit(0);

    /* Allocate the Argv list */
  if(!(Argv = (char **)malloc(sizeof(char *)*(argc+1))))
  { fprintf(stderr,
       "Memory Allocation Error : Unable to allocate argument list\n");
    exit(1);
  }
  *Argvp = Argv;
  
  j = 0;
  i = 1;
  getpars = 1; /* Set to 0 when pars must be added to the argument list */

  /* Set all the default values */
  if(!e_parameter_Set_Default_Value(Par)) return(-1);

  while(i < argc) /* Scan the arguments list */
  { s = argv[i];
   
    if (getpars && (s[0] == '-')) 
    { if(s[1] == 0) /* - without name: */
      { if(mode & E_PARAMETER_MINUSOK)
        { Argv[j++] = s;
        }
        else if (mode & E_PARAMETER_MINUSLAST)
        { getpars = 0;
        }
        else goto InvPar;
      }

      else /* Par */
      { if(!(par = e_parameter_Get_Par_Details(Par,s+1)))
        { if(mode & E_PARAMETER_BADPAROK) 
          { /* Add par to the argument list */
            Argv[j++] = s;
          }
          else goto InvPar;
        }
        else 
        switch (par->type)
        { case E_CTYPE_INTEGER :
                  i++; 
                  if(i >= argc) goto NoValue;
                  sscanf(argv[i],"%d",(int *)par->p);
                  par->Read = 1;
                  break;

          case E_CTYPE_REAL :
                  i++; 
                  if(i >= argc) goto NoValue; 
                  sscanf(argv[i],"%lg",(double *)par->p);
                  par->Read = 1;
                  break;

          case E_CTYPE_CHAR :
                  i++; 
                  if(i >= argc) goto NoValue; 
                  sscanf(argv[i],"%c",(char *)par->p);
                  par->Read = 1;
                  break;

          case E_CTYPE_STRING :
                  i++; 
                  if(i >= argc) goto NoValue;
                  strcpy(par->p,argv[i]);
                  par->Read = 1;
                  break;

          case E_CTYPE_BOOLEAN :
                  i++; 
                  if(i >= argc) goto NoValue;
                  sscanf(argv[i],"%d",(int *)par->p);
                  if(*((int *)par->p)) *((int *)par->p) = 1;
                  par->Read = 1;
                  break;

          case E_CTYPE_FLAG :
                  *((int *)par->p) = 1;
                  par->Read = 1;
                  break;

          default : 
                  Argv[j++] = s;
                  break;
        }
      }
    }
    else 
    { Argv[j++] = s;
    }
    i++;
  }

  Argv[j] = 0;
  if(!e_parameter_Check_Par_Read(Par)) return(-1);

  return(j);

NoValue:
  e_error_syntax("No Value for : %s",s);
  return(-1);

InvPar:
  e_error_syntax("Invalid parameter : %s",s,0);
  return(-1);

} 

/****************************************************/
/* Initialise each parameter with the default value */
/* (when there is one).                             */
/* The value of read is set to 1 for each parameter */
/* with a default value                             */
/* Return : 1 : OK                                  */
/*          0 : error (invalid type)                */
/****************************************************/
int e_parameter_Set_Default_Value(e_parameter *pars)
{ 
  /* The last element of pars is a dummy line with type NULL */ 
  while(pars->type)
  { /* By default Flags are always off */
    if(pars->type == E_CTYPE_FLAG) 
    { pars->Read = 1;
      *((int *)pars->p) = 0;
    }
    else
    { switch(pars->type)
      { case E_CTYPE_INTEGER:
              if(pars->init_val)
              { sscanf(pars->init_val,"%d",(int *)pars->p);
                pars->Read = 1;
              }
                    break;
        case E_CTYPE_REAL:
              if(pars->init_val)
              { sscanf(pars->init_val,"%lg",(double *)pars->p);
                pars->Read = 1;
              }
                    break;
        case E_CTYPE_CHAR:
              if(pars->init_val)
              { sscanf(pars->init_val,"%c",(char *)pars->p);
                pars->Read = 1;
              }
                    break;
        case E_CTYPE_STRING:
              if(pars->init_val)
              { strcpy(pars->p,pars->init_val);
                pars->Read = 1;
              }
              else ((char *)pars->p)[0] = 0; /* Allways set a string to 0 */

                    break;
        case E_CTYPE_BOOLEAN:
              if(pars->init_val)
              { if(!strcmp(pars->init_val,"1") || 
                   !strcmp(pars->init_val,"true") ||
                   !strcmp(pars->init_val,"TRUE") || 
                   !strcmp(pars->init_val,"True") 
                  ) *((int *)pars->p) = 1;
                else if(!strcmp(pars->init_val,"0") || 
                   !strcmp(pars->init_val,"false") ||
                   !strcmp(pars->init_val,"FALSE") || 
                   !strcmp(pars->init_val,"False") 
                  ) *((int *)pars->p) = 0;
                else
                { e_error_syntax(
               "Set_Default_Value : Invalid boolean default value %s for %s", 
                                  pars->init_val,pars->name);
                   return(0);
                }
              }
                    break;
        default : e_error_syntax(
               "Set_Default_Value : Invalid parameter type for parameter %s", 
                                  pars->init_val,pars->name);
          return(0);
      }
    }
    pars++;
  }
  return(1);
}

/****************************************************************/ 
/* Search the parameter entry associated to name in Parameter g */
/* p : Pointer to an array of pointer of Parameters             */
/* Return : pointer to the Parameter                            */
/*          0 : not found                                       */
/****************************************************************/
e_parameter *e_parameter_Get_Par_Details(e_parameter *pars,char *name)
{
       /* Scan all the parameters in the table */
  while(pars->type) /* Last element of p is a dummy line with type NULL */
  {    /* when the parameter is found: return a pointer to its def. line */
    if(!strcmp(pars->name,name)) return(pars);
    if(pars->alias1 && !strcmp(pars->alias1,name)) return(pars);
    if(pars->alias2 && !strcmp(pars->alias2,name)) return(pars);
    pars++;
  }
  return(0); /* No match found */
}

/****************************************************/
/* Check that all the parameters have been read.    */
/* p : Pointer to an array of pointer of Parameters */
/* Prints the name of all unread parameters.        */
/* Return : 1 : OK                                  */
/*          0 : some parameters are not set.        */
/****************************************************/
int e_parameter_Check_Par_Read(e_parameter *pars)
{ int r;
  char Buff[1024];
 
  r = 1;
 
  Buff[0] = 0;

      /* The last element of g is a dummy line with type NULL */ 
  while(pars->type)
  { if(!pars->Read) /* look for unread pars */
    { /* Merge unset parameters */ 
      strcat(Buff,pars->name); 
      strcat(Buff," ");
      r = 0;
    }
    pars++;
  }

  if(!r)
  { e_error_syntax("The following parameters have not been set : %s",Buff);
  }

  return(r);
}

/**********************************************************/
/* Print the array of characters on stderr                */
/* p : pointer to an array of characters (lines to print) */
/*     the last line must be a null pointer.              */
/* return : 1                                             */
/**********************************************************/
int e_Syntax(char **p)
{
  while(*p) /* Scan all the elements of p until a NULL entry is found! */
  { fprintf(stderr,"%s\n",*p++);
  }
  return(1);
}
