#include <easy.h>

e_obj_descriptor e_obj_descriptor_string = {
  "string",           /* Name      */
  0,                  /* def_mode  */
  malloc,             /* _alloc    */
  free,               /* _free     */
  realloc,            /* _realloc  */
  e_string_free,      /* free      */
  e_string_to_string, /* to_string */
  e_string_clone,     /* clone     */
  e_string_equals,    /* equals    */
};

/********************************************************************/
/* e_string : typical usage                                          */
/********************************************************************/
/* A : dynamic allocation                                           */
/********************************************************************/
/* e_string *stp;                                                    */
/* char c,*s;                                                       */
/* int pos,i;                                                       */
/*  if(!(stp = e_string_new(0,"Just a string!",0))) goto error;     */
/*  pos = 0;                                                        */
/*  while((c = e_string_get(stp,pos++)) != EOF) { ... }             */
/*                                                                  */ 
/*  e_string_set_index(stp,0);                                      */
/*  while((c = e_string_iget(stp)) != EOF) { ... }                  */
/*                                                                  */ 
/*  s = "Something else"; i = pos =0;                               */
/*  while(s[i]) { e_string_put(stp,s[i++],pos++); }                 */
/*                                                                  */ 
/*  e_string_set_index(stp,0);                                      */
/*  s = "Something else"; i = 0;                                    */
/*  while(s[i]) { e_string_iput(stp,s[i++]); }                      */
/*                                                                  */ 
/*  e_string_free(stp);                                             */
/********************************************************************/
/* B : static allocation                                            */
/********************************************************************/
/* e_string stp;                                                    */
/* char c,*s;                                                       */
/* int pos,i;                                                       */
/*  if(!e_string_new(stp,"Just a string!",0)) goto error;           */
/*  pos = 0;                                                        */
/*  while((c = e_string_get(&stp,pos++)) != EOF) { ... }            */
/*  e_string_free(&stp);                                            */
/********************************************************************/

/***************************/
/* Initialisation function */
/***************************/
int e_string_init(void *p)
{ /* Nothing to do here */
  return(1);
}

/********************************************************************/
/* Initialise or Create a new e_string                              */
/* stp : pointer to new e_string if NULL, use malloc to allocate it */
/* s   : initial value of the string. If NULL use empty string      */
/* size: initial size for the string. If NULL or less than size of  */ 
/*       s, use the lenfth of s + 1                                 */
/* Return : pointer to e_string.                                    */
/*          0 : error                                               */
/********************************************************************/
e_string *e_string_new(e_string *stp,char *s, int size)
{ 
                   /* initialise with empty string by default */ 
  if (!s) s = "";  

                   /* alloc stp is not provided */
  if (!stp)
  { if (!(stp = e_obj_descriptor_string._alloc(sizeof(e_string))))
         return((e_string *)e_error_no_memory("e_string_new"));
    stp->h.refs = 1;
  }
  else stp->h.refs = -1;
  
  stp->h.obj_desc_p = &e_obj_descriptor_string;
  stp->h.mode = 0;

                   /* default size is string size */
  stp->len = strlen(s);
  if(size <= stp->len) size = stp->len+1;     

                   /* Allocate string buffer and set buffer size */
  if (!(stp->s = e_obj_descriptor_string._alloc(size)))
  {                /* free stp if is was mallocated */
    e_error_no_memory("e_string_new");
    goto error;
  }

  stp->size = size;
  strncpy(stp->s,s,stp->len+1);
  stp->index = 0;
   
  return(stp);

error:
  if(stp && (stp->h.refs > 0)) e_obj_descriptor_string._free(stp); 
  return(0);
}

/**********************************************************/
/* Make sure the string is at least of the specified size */
/* Realocate if it is not the case.                       */
/* The pointer s might change as a result.                */
/* stp : the string, assumed to be of the correct type    */
/* Return : 1 : OK                                        */
/*          0 : error                                     */
/**********************************************************/
int e_string_resize(e_string *stp,int size)
{ void *p;

  if(!stp) return(0);

  size++; /* allow for 0 at end of char */  

  if(size > stp->size)
  { if(!(p = e_obj_descriptor_string._realloc(stp->s,size))) 
         return(e_error_no_memory("e_string_new"));
    stp->s = p;
    stp->size = size;
  }

  return(1);
}

/***************************************************************************/
/* Free the string                                                         */
/* If the refs counter is -1 : static string : free the string buffer only */
/* If the refs counter is 1 : free the allocated memory for the e_string   */
/* Otherwise decrement the counter by 1.                                   */
/* stp : the string, assumed to be of the correct type                     */
/* REF : stp : -1                                                          */
/***************************************************************************/
void e_string_free(void *stp)
{ if(stp)
  { if(E_OBJ_REF(stp) == 0)
    { e_error(E_ERROR_INCONSISTENCY,E_ERROR_LEVEL_PANIC,
            "Trying to free a free string (%x) : \"%s\"",
             stp,E_STRING_S(stp));
      e_error_flush(E_ERROR_LEVEL_DEBUG); 
      return;
    }
    if (E_OBJ_REF(stp) < 0)      /* static string, free the buffer */
    { if (E_STRING_S(stp)) e_obj_descriptor_string._free(E_STRING_S(stp));
      E_STRING_S(stp) = 0;
      return;
    }
    if (E_OBJ(stp)->refs-- > 1) return;   /* decrement counter    */

                                    /* Last reference, 
                                       free the string buffer and the string */
    if (E_STRING_S(stp)) e_obj_descriptor_string._free(E_STRING_S(stp));
    e_obj_descriptor_string._free(stp);
  }
}

/*******************************************************/
/* Convert the string to a string                      */
/* If require, tab and quote the string                */
/* Return : new string                                 */
/*          0 : error                                  */
/* stp : the string, assumed to be of the correct type */
/* LINK : stp : 0                                      */
/*        RETURN : +1                                  */
/*******************************************************/
void *e_string_to_string(void *stp,int tab,e_flag mode)
{ 
  if(!stp) /* Null pointer -> return empty string */
  { return(e_string_new(0,"",1));
  }

  E_USE(stp);
  if(mode & E_TO_STR_QUOTE)
  { /* Quote the string */
    if(!(stp = e_string_quote(stp))) return(0);
  }

  if(!(mode & E_TO_STR_NO_TAB))
  { /* add tab to the string */
    stp = e_string_insert_chars(stp,0,tab,' ');
  }

  return(stp);
}

/*********************************************/
/* Clone a string                            */
/* mode : cloning mode                       */
/* If the string can't be cloned, return stp */
/* and increase its ref counter              */
/* stp : the string to clone, assumed to be  */
/*       of the correct type                 */
/* Return : cloned e_string                  */
/*          0 : Error                        */
/* LINK : stp : 0                            */
/*        RETURN : +1                        */
/*********************************************/
void *e_string_clone(void *stp,e_flag mode)
{ e_string *news;

         /* Cloning a null pointer : returns NULL */
  if(!stp) return(0); 

                 /* Can't clone : return the same object */
  if((E_OBJ_NO_CLONE(stp))) return(E_USE(stp));

  if((news = e_string_new(0,E_STRING_S(stp),0)))
  { news->index = E_STRING_INDEX(stp);
  }
  return(news);
}

/*********************************************/
/* Test if the 2 strings are equals          */
/* stp : the string to clone, assumed to be  */
/*       of the correct type                 */
/* Return : 0 : p1 and p2 are different      */
/*          1 : p1 and p2 are equal          */
/* REF : p1,p2 : 0                           */
/*********************************************/
int e_string_equals(void *p1,void *p2)
{ 
  if(p1 == p2) return(1);

  if(!p1 || !p2) return(0); /* p1 or p2 is NULL -> they differ */

  /* p1 and p2 are of different type => they differ */
  if(E_OBJ_DESCR(p1) != E_OBJ_DESCR(p2)) return(0);

  return(strcmp(E_STRING_S(p1),E_STRING_S(p2)) ? 0 : 1);
}

/********************************************************************/
/* Set the e_string index pointer                                   */
/* index : index position in the e_string.                          */
/*     if index < 0  count from end of string : -1 is the last char */
/*     len+1-index must be positive otherwise index is set to 0.    */
/* Link : stp : 0                                                   */
/********************************************************************/
void e_string_set_index(e_string *stp, int index)
{ 
  E_NULL_ERROR(e_check_type(stp,E_TYPE_STRING,"e_string_set_index(stp,-)"));

  if (index < 0)
  { index = stp->len + 1 - index;
    if (index < 0) index = 0;
  }

  stp->index = index;

error:
}

/********************************************************************/
/* Get char at the given position in the string.                    */
/* The index is not changed                                         */
/* To hcek if an error occur, do the folowing test:                 */
/*   if(E_TEST_ERROR(E_ERROR_ANY)) }{...}                           */
/* stp : pointer to e_string.                                       */
/* pos : index position for the character.                          */
/*       if pos < 0  count from end of string : -1 is the last char */
/*       len+1-pos must be positive.                                */
/* Return char at pos                                               */
/*        EOF : end of string reached.                              */  
/*        0 : ERROR or valid NULL char                              */
/* Link : stp : 0                                                   */
/********************************************************************/
int e_string_get(e_string *stp,int pos)
{ 
  E_NULL_ERROR(e_check_type(stp,E_TYPE_STRING,"e_string_get(stp,-)"));
  E_NULL_ERROR(e_error_is_readable(stp,"e_string_get"));

  if (pos < 0)
  { pos = stp->len + 1 - pos;
    if (pos < 0) return(0);
  }

  if(pos >= stp->len) return(EOF);

  return(stp->s[pos]);

error:
  return(0);  
}

/********************************************************************/
/* Put char at the given position in the string.                    */
/* The buffer is increased if necessary and the characters between  */ 
/* the old end of string and pos is filled up with SPACES.          */
/* The index is not changed                                         */
/* stp : pointer to e_string.                                       */
/* c   : character to insert                                        */
/* pos : index position for the new character.                      */
/*       if pos < 0  count from end of string : -1 is the last char */
/*       len+1-pos must be positive.                                */
/* Return 1 : OK                                                    */
/*        0 : ERROR                                                 */
/* Link : stp : 0                                                   */
/********************************************************************/
int e_string_put(e_string *stp,char c, int pos)
{ int i;

  E_NULL_ERROR(e_check_type(stp,E_TYPE_STRING,"e_string_put(stp,-)"));
  E_NULL_ERROR(e_error_is_writable(stp,"e_string_put"));

  if (pos < 0)
  { pos = stp->len + 1 - pos;
    if (pos < 0) return(0);
  }

                        /* Make sure the string is large enough */
  if(stp->size <= pos)  
  { E_NULL_ERROR(e_string_resize(stp,pos+1));
  }
                       /* fill gap with spaces */
  if (pos > stp->len)
  { for(i = stp->len; i < pos; i++)
    { stp->s[i] = ' ';
    }
  }

  stp->s[pos] = c;
  stp->s[pos+1] = 0;

  if(pos >= stp->len) stp->len = pos+1;

  return(1);

error:
  return(0);  
}

/*******************************/
/* Append 1 char to the string */
/* The index is not changed    */
/* Return 1 : OK               */
/*        0 : Error            */
/* Link : stp : 0              */
/*******************************/
int e_string_append_char(e_string *stp,char c)
{ 
  if(!stp) return(e_error_null_pointer("e_string_append_char"));

  return(e_string_put(stp,c, stp->len));

}

/****************************************/
/* Get the next char in the string      */
/* Return : next char                   */
/*          EOF : end of string reached */
/*          0 : error                   */
/* Link : stp : 0                       */
/****************************************/
int e_string_iget(e_string *stp)
{ 
  E_NULL_ERROR(e_check_type(stp,E_TYPE_STRING,"e_string_iget(stp)"));
  E_NULL_ERROR(e_error_is_readable(stp,"e_string_iget"));

  if(stp->index >= stp->len) return(EOF);

  return(stp->s[stp->index++]);

error:
  return(0); 
}

/********************************************/
/* Put char in the string at index position */
/* Return : 1 OK                            */
/*          0 : error                       */
/* Link : stp : 0                           */
/********************************************/
int e_string_iput(e_string *stp,char c)
{ 
  E_NULL_ERROR(e_check_type(stp,E_TYPE_STRING,"e_string_iput(stp)"));
  E_NULL_ERROR(e_string_put(stp,c,stp->index));

  stp->index++;
  return(1);

error:
  return(0); 
}

/*********************************************/
/* Blank the string : set len and index to 0 */
/* stp : string to blank                     */
/* Link : stp : -1                           */
/*        Ret : +1                           */
/*********************************************/
e_string *e_string_blank(e_string *stp)
{ 
  E_NULL_ERROR(e_check_type(stp,E_TYPE_STRING,
                            "e_string_blank(stp)"));

  stp->s[0] = 0;
  stp->len = 0;
  stp->index = 0;

  return(stp);

error:
  if(stp)  E_FREE(stp);
  return(0);   
}

/*********************************************/
/* Reset the string to the string value str  */
/* Ensure that the string is large enough.   */
/* Reset the index to 0                      */
/* stp : string to blank                     */
/* Link : stp : -1                           */
/*        Ret : +1                           */
/*********************************************/
e_string *e_string_reset(e_string *stp,char *str)
{ int n;
  
  E_NULL_ERROR(e_check_type(stp,E_TYPE_STRING,
                            "e_string_reset(stp,-)"));
  if(!stp) 
  { e_error_null_pointer("e_string_reset(-str)");
    goto error;
  }

  n = strlen(str);
  if(!e_string_resize(stp,n+1)) goto error;
  strcpy(E_STRING_S(stp),str);

  stp->len = n;
  stp->index = 0;

  return(stp);

error:
  if(stp)  E_FREE(stp);
  return(0);   
}

/**************************************************************************/
/* Insert n characters starting at pos in stp.                            */
/* If c is NULL, the inserted block is left non initialised.              */
/* The string is shifted properly.                                        */
/* If the Ref counter is 1 and the object is big enough, stp is modified  */
/* otherwise, a new object is created and returned                        */
/* stp : string to modify: assumed to be of the correct type              */
/* Return : modified string                                               */
/*          0 : Error                                                     */
/* LINK stp    : -1                                                       */
/*      Return : +1                                                       */
/**************************************************************************/
e_string *e_string_insert_chars(e_string *stp,int pos, int n,char c)
{ int newlen,i,j;
  e_string *nstp = 0;
  char *s,*d;

  if(!stp) return((e_string *)e_error_null_pointer("e_string_insert_chars"));
 
  if(n <= 0) return(stp);

  newlen = E_MAX(stp->len,pos) +n;

  if((E_OBJ_REF(stp) > 1) || (newlen >= stp->size) || E_OBJ_READ_ONLY(stp))
  {                       /* create a new string */
    E_NULL_ERROR(nstp = e_string_new(0,0,newlen+1));

    s = E_STRING_S(stp);
    d = E_STRING_S(nstp);
    
    j = 0;                
    if (pos > stp->len)
    {                     /* copy the string */ 
      for(i=0; i < stp->len; i++) d[j++] = s[i];
                          /* Fill up with spaces */ 
      for(i=stp->len; i < pos; i++) d[j++] = ' '; 

    }
    else                  /* copy beginning of string */ 
    { for(i=0; i < pos; i++) d[j++] = s[i];
    }
 
    if(c)                 /* write c n times          */
    { while(j < (pos+n))   d[j++] = c;
    }
    j = pos+n;            /* copy end of string */ 
    for(i=pos; i < stp->len; i++) d[j++] = s[i];

    E_FREE(stp);          /* derefer stp */ 
    stp = nstp; 
    nstp = 0;
    s = E_STRING_S(stp);
  } 
  else
  { if(newlen >= stp->size) 
    { E_NULL_ERROR(e_string_resize(stp,newlen)); 
    }

    s = E_STRING_S(stp);
    
                          /* shift the string */                      
    for(i=stp->len; i >= pos; i--) s[i+n] = s[i];
    
                          /* pos > len : fill gap with spaces */
    for(i=stp->len; i < pos; i++) s[i] = ' ';

                          /* write c if it is non null */                      
    if(c)
    { for(i=pos; i < n; i++) s[i] = c;
    } 
  } 

  stp->len = newlen;
  s[newlen] = 0;

  return(stp);


error:
  if(stp)  E_FREE(stp);
  if(nstp) E_FREE(nstp);
  return(0);  
}

/*********************************************/
/* Insert string stp2 at pos into string stp */
/* stp : string to insert into               */
/* stp2 : string to insert                   */
/* pos : insert position                     */
/* Link : stp : -1                           */
/*        stp2 : -1                          */
/*        Ret : +1                           */
/*********************************************/
e_string *e_string_insert_string(e_string *stp,e_string *stp2,int pos)
{ int n;

  E_NULL_ERROR(e_check_type(stp,E_TYPE_STRING,
                            "e_string_insert_string(stp,-)"));
  E_NULL_ERROR(e_check_type(stp2,E_TYPE_STRING,
                            "e_string_insert_string(-,stp2)"));

  n = stp2->len;
    
  E_NULL_ERROR(stp = e_string_insert_chars(stp,pos, n,0));

  memcpy(E_STRING_S(stp)+pos,E_STRING_S(stp2),n);

  E_FREE(stp2);

  return(stp);

error:
  if(stp)  E_FREE(stp);
  if(stp2) E_FREE(stp2);
  return(0);   
}

/*********************************************/
/* Delete n chars from stp at pos            */
/* stp : string to insert into               */
/* pos : delete from this character          */
/* n   : number of char to delete            */
/* Link : stp : -1                           */
/*        Ret : +1                           */
/*********************************************/
e_string *e_string_delete(e_string *stp,int pos,int n)
{ int newlen;
  e_string *nstp = 0;

  E_NULL_ERROR(e_check_type(stp,E_TYPE_STRING,"e_string_delete(stp,-,-)"));

  if(pos >= stp->len)      /* nothing to remove */
  { return(stp);
  }
  
  if(pos+n >= stp->len) newlen = pos-1;
  else newlen = stp->len-n;

  if((E_OBJ_REF(stp) > 1) || E_OBJ_READ_ONLY(stp))
  { if(!(nstp = e_string_new(0,0,newlen))) goto error;

                            /* copy beginning of string */
    memcpy(E_STRING_S(nstp),E_STRING_S(stp),pos);

    if((pos+n) < stp->len)  /* copy end of string       */
    { memcpy(E_STRING_S(nstp)+pos,E_STRING_S(stp)+(pos+n),stp->len-pos-n);
    }
    E_STRING_S(nstp)[newlen] = 0;

    E_FREE(stp);
    stp = nstp;
  }
  else
  { if((pos+n) < stp->len)  /* move end of string       */
    { memcpy(E_STRING_S(stp)+pos,E_STRING_S(stp)+(pos+n),stp->len-pos-n);
    }
    E_STRING_S(stp)[newlen] = 0;
  }

  return(stp);

error:
  if(stp) E_FREE(stp);
  return(0);   
}

/************************************/
/* Create a quoted strings from stp */
/* Return : new string              */
/*          0 : error               */
/* LINK : stp : -1                  */
/*        Ret : +1                  */
/************************************/
e_string *e_string_quote(e_string *stp)
{ int len,i,crit_len;
  char *s1,*s2,c,c2;
  e_string *nstp = 0;


  E_NULL_ERROR(e_check_type(stp,E_TYPE_STRING,"e_string_quote(stp)"));
  E_NULL_ERROR(e_error_is_readable(stp,"e_string_quote"));

         /* New string 1.1 larger is reasonable */
  len =(int)(stp->len*1.1);
  if(len < 20) len = 20; 

  E_NULL_ERROR(nstp = e_string_new(0,0,len));

  s1 = E_STRING_S(stp);
  s2 = E_STRING_S(nstp);

  crit_len = nstp->size-4;
  i = 0;

  s2[i++] = '"';

  while((c = *s1++))
  { c2 = 0;
    if(c < ' ')        /* special chars */
    { c = '\\';
      switch (c)
      { case 7  : c2 = 'a'; break;
        case 8  : c2 = 'b'; break;
        case 9  : c2 = 't'; break;
        case 10 : c2 = 'n'; break;
        case 11 : c2 = 'v'; break;
        case 12 : c2 = 'f'; break;
        case 13 : c2 = 'r'; break;
        default : c = '.'; break;
      }
    }
                     /* escape quote */
    else if(c == '"') { c = '\\'; c2 = c; }

    if(i >= crit_len) 
    { if(!e_string_resize(nstp,(int)(i*1.1+20)))
      { e_error(E_ERROR_MEMORY,E_ERROR_LEVEL_USER,
              "Unable to quote string (%x) : \"%s\"",E_STRING_S(stp),stp);
        goto error;
      }  
      crit_len = nstp->size-4;
      s2 = E_STRING_S(nstp);
    }
    s2[i++] = c;
    if(c2) s2[i++] = c2;
  }

  s2[i++] = '"';
  s2[i] = 0;
  nstp->len = i;

  E_FREE(stp);

  return(nstp);

error:
  if(stp) E_FREE(stp);
  if(nstp) E_FREE(nstp);
  return(0);
}

/****************************************************/
/* Append the string buffer str to the e_string stp */
/* Return : new string                              */
/*          0 : error                               */
/* Link : stp -1                                    */
/*        Ret + 1                                   */
/****************************************************/
e_string *e_string_append_buffer(e_string *stp,char *str)
{ int n,pos;

  E_NULL_ERROR(e_check_type(stp,E_TYPE_STRING,
                            "e_string_append_buffer(stp,-,-)"));

  n = strlen(str);
  pos = E_STRING_LEN(stp);
  
  E_NULL_ERROR(stp = e_string_insert_chars(stp,pos,n,0));

  strcpy(E_STRING_S(stp)+pos,str);

  return(stp);

error:
  if(stp)  E_FREE(stp);
  return(0);   
}

/*********************************************/
/* Append e_string stp_2 to the e_string stp */
/* Return : new string                       */
/*          0 : error                        */
/* Link : stp  -1                            */
/*        stp2 -1                            */
/*        Ret + 1                            */
/*********************************************/
e_string *e_string_append_string(e_string *stp,e_string *stp2)
{ 
  if(!stp) 
  { e_error_null_pointer("e_string_append_string(stp,)"); 
    goto error;
  }
  E_NULL_ERROR(e_check_type(stp2,E_TYPE_STRING,
                            "e_string_append_string(-,stp2)"));

  stp = e_string_append_buffer(stp,E_STRING_S(stp2));

  E_FREE(stp2);

  return(stp);

error:
  if(stp)  E_FREE(stp);
  return(0);   
}
